@@ -27,9 +27,9 @@ struct hardlink_info
 extern List *hardlist;
 extern char *working_dir;
 
-Node *lookup_file_by_inode PROTO ((const char *));
-void update_hardlink_info PROTO ((const char *));
-List *list_linked_files_on_disk PROTO ((char *));
-int compare_linkage_lists PROTO ((List *, List *));
-int find_checkedout_proc PROTO ((Node *, void *));
+Node *lookup_file_by_inode (const char *);
+void update_hardlink_info (const char *);
+List *list_linked_files_on_disk (char *);
+int compare_linkage_lists (List *, List *);
+int find_checkedout_proc (Node *, void *);
 #endif /* PRESERVE_PERMISSIONS_SUPPORT */
