@@ -49,16 +49,18 @@ struct list
 typedef struct list List;
 
 List *getlist (void);
-Node *findnode (List * list, const char *key);
-Node *findnode_fn (List * list, const char *key);
+Node *findnode (List *list, const char *key);
+Node *findnode_fn (List *list, const char *key);
 Node *getnode (void);
-int insert_before (List * list, Node * marker, Node * p);
-int addnode (List * list, Node * p);
-int addnode_at_front (List * list, Node * p);
-int walklist (List * list, int (*)(Node *n, void *closure), void *closure);
+int insert_before (List *list, Node *marker, Node *p);
+int addnode (List *list, Node *p);
+int addnode_at_front (List *list, Node *p);
+int walklist (List *list, int (*)(Node *n, void *closure), void *closure);
 int list_isempty (List *list);
-void dellist (List ** listp);
-void delnode (Node * p);
-void freenode (Node * p);
-void sortlist (List * list, int (*)(const Node *, const Node *));
-int fsortcmp (const Node * p, const Node * q);
+void removenode (Node *p);
+void mergelists (List *dest, List **src);
+void dellist (List **listp);
+void delnode (Node *p);
+void freenode (Node *p);
+void sortlist (List *list, int (*)(const Node *, const Node *));
+int fsortcmp (const Node *p, const Node *q);
