@@ -1,6 +1,11 @@
 /*
- * Copyright (c) 1992, Brian Berliner and Jeff Polk
- * Copyright (c) 1989-1992, Brian Berliner
+ * Copyright (C) 1986-2005 The Free Software Foundation, Inc.
+ *
+ * Portions Copyright (C) 1998-2005 Derek Price, Ximbiot <http://ximbiot.com>,
+ *                                  and others.
+ *
+ * Portions Copyright (C) 1992, Brian Berliner and Jeff Polk
+ * Portions Copyright (C) 1989-1992, Brian Berliner
  *
  * You may distribute under the terms of the GNU General Public License as
  * specified in the README file that comes with the CVS source distribution.
@@ -34,6 +39,7 @@
  */
 
 #include "cvs.h"
+#include <assert.h>
 #include "save-cwd.h"
 #ifdef SERVER_SUPPORT
 # include "md5.h"
@@ -91,8 +97,8 @@ static int warned;
    the setting.  See update_dirent_proc.  */
 static char *tag_update_dir;
 
-static char *join_rev1, *date_rev1;
-static char *join_rev2, *date_rev2;
+static char *join_rev1, *join_date1;
+static char *join_rev2, *join_date2;
 static int aflag = 0;
 static int toss_local_changes = 0;
 static int force_tag_match = 1;
@@ -104,7 +110,7 @@ static int dotemplate = 0;
 static int patches = 0;
 static int rcs_diff_patches = 0;
 #endif
-static List *ignlist = (List *) NULL;
+static List *ignlist = NULL;
 static time_t last_register_time;
 static const char *const update_usage[] =
 {
@@ -139,10 +145,16 @@ update (int argc, char **argv)
     int c, err;
     int local = 0;			/* recursive by default */
     int which;				/* where to look for files and dirs */
+    char *xjoin_rev1, *xjoin_date1,
+	 *xjoin_rev2, *xjoin_date2,
+	 *join_orig1, *join_orig2;
 
     if (argc == -1)
 	usage (update_usage);
 
+    xjoin_rev1 = xjoin_date1 = xjoin_rev2 = xjoin_date2 = join_orig1 =
+	         join_orig2 = NULL;
+
     ign_setup ();
     wrap_setup ();
 
@@ -177,11 +189,9 @@ update (int argc, char **argv)
 		break;
 	    case 'Q':
 	    case 'q':
-#ifdef SERVER_SUPPORT
 		/* The CVS 1.5 client sends these options (in addition to
 		   Global_option requests), so we must ignore them.  */
 		if (!server_active)
-#endif
 		    error (1, 0,
 			   "-q or -Q must be specified before \"%s\"",
 			   cvs_cmd_name);
@@ -193,9 +203,10 @@ update (int argc, char **argv)
 		force_tag_match = 0;
 		break;
 	    case 'r':
-		tag = optarg;
+		parse_tagdate (&tag, &date, optarg);
 		break;
 	    case 'D':
+		if (date) free (date);
 		date = Make_Date (optarg);
 		break;
 	    case 'P':
@@ -206,12 +217,18 @@ update (int argc, char **argv)
 		noexec = 1;		/* so no locks will be created */
 		break;
 	    case 'j':
-		if (join_rev2)
+		if (join_orig2)
 		    error (1, 0, "only two -j options can be specified");
-		if (join_rev1)
-		    join_rev2 = optarg;
+		if (join_orig1)
+		{
+		    join_orig2 = xstrdup (optarg);
+		    parse_tagdate (&xjoin_rev2, &xjoin_date2, optarg);
+		}
 		else
-		    join_rev1 = optarg;
+		{
+		    join_orig1 = xstrdup (optarg);
+		    parse_tagdate (&xjoin_rev1, &xjoin_date1, optarg);
+		}
 		break;
 	    case 'u':
 #ifdef SERVER_SUPPORT
@@ -268,10 +285,10 @@ update (int argc, char **argv)
 		send_arg (options);
 	    if (date)
 		client_senddate (date);
-	    if (join_rev1)
-		option_with_arg ("-j", join_rev1);
-	    if (join_rev2)
-		option_with_arg ("-j", join_rev2);
+	    if (join_orig1)
+		option_with_arg ("-j", join_orig1);
+	    if (join_orig2)
+		option_with_arg ("-j", join_orig2);
 	    wrap_send ();
 
 	    if (failed_patches_count == 0)
@@ -368,9 +385,9 @@ update (int argc, char **argv)
     if (tag != NULL)
 	tag_check_valid (tag, argc, argv, local, aflag, "", false);
     if (join_rev1 != NULL)
-        tag_check_valid_join (join_rev1, argc, argv, local, aflag, "");
+	tag_check_valid (xjoin_rev1, argc, argv, local, aflag, "", false);
     if (join_rev2 != NULL)
-        tag_check_valid_join (join_rev2, argc, argv, local, aflag, "");
+	tag_check_valid (xjoin_rev2, argc, argv, local, aflag, "", false);
 
     /*
      * If we are updating the entire directory (for real) and building dirs
@@ -409,17 +426,21 @@ update (int argc, char **argv)
     which = W_LOCAL | W_REPOS;
 
     /* look in the attic too if a tag or date is specified */
-    if (tag != NULL || date != NULL || joining())
+    if (tag || date || join_orig1)
+    {
+	TRACE (TRACE_DATA, "update: searching attic");
 	which |= W_ATTIC;
+    }
 
     /* call the command line interface */
     err = do_update (argc, argv, options, tag, date, force_tag_match,
 		     local, update_build_dirs, aflag, update_prune_dirs,
-		     pipeout, which, join_rev1, join_rev2, NULL, 1, NULL);
+		     pipeout, which, xjoin_rev1, xjoin_date1, xjoin_rev2,
+		     xjoin_date2, NULL, 1, NULL);
 
-    /* free the space Make_Date allocated if necessary */
-    if (date != NULL)
-	free (date);
+    /* Free the space allocated for tags and dates, if necessary.  */
+    if (tag) free (tag);
+    if (date) free (date);
 
     return err;
 }
@@ -437,11 +458,22 @@ update (int argc, char **argv)
 int
 do_update (int argc, char **argv, char *xoptions, char *xtag, char *xdate,
            int xforce, int local, int xbuild, int xaflag, int xprune,
-           int xpipeout, int which, char *xjoin_rev1, char *xjoin_rev2,
+           int xpipeout, int which, char *xjoin_rev1, char *xjoin_date1,
+	   char *xjoin_rev2, char *xjoin_date2,
            char *preload_update_dir, int xdotemplate, char *repository)
 {
     int err = 0;
-    char *cp;
+
+    TRACE (TRACE_FUNCTION,
+"do_update (%s, %s, %s, %d, %d, %d, %d, %d, %d, %d, %s, %s, %s, %s, %s, %d, %s)",
+           xoptions ? xoptions : "(null)", xtag ? xtag : "(null)",
+	   xdate ? xdate : "(null)", xforce, local, xbuild, xaflag, xprune,
+	   xpipeout, which, xjoin_rev1 ? xjoin_rev1 : "(null)",
+	   xjoin_date1 ? xjoin_date1 : "(null)",
+	   xjoin_rev2 ? xjoin_rev2 : "(null)",
+	   xjoin_date2 ? xjoin_date2 : "(null)",
+	   preload_update_dir ? preload_update_dir : "(null)", xdotemplate,
+	   repository ? repository : "(null)");
 
     /* fill in the statics */
     options = xoptions;
@@ -456,21 +488,9 @@ do_update (int argc, char **argv, char *xoptions, char *xtag, char *xdate,
 
     /* setup the join support */
     join_rev1 = xjoin_rev1;
+    join_date1 = xjoin_date1;
     join_rev2 = xjoin_rev2;
-    if (join_rev1 && (cp = strchr (join_rev1, ':')) != NULL)
-    {
-	*cp++ = '\0';
-	date_rev1 = Make_Date (cp);
-    }
-    else
-	date_rev1 = NULL;
-    if (join_rev2 && (cp = strchr (join_rev2, ':')) != NULL)
-    {
-	*cp++ = '\0';
-	date_rev2 = Make_Date (cp);
-    }
-    else
-	date_rev2 = NULL;
+    join_date2 = xjoin_date2;
 
 #ifdef PRESERVE_PERMISSIONS_SUPPORT
     if (preserve_perms)
@@ -504,13 +524,8 @@ do_update (int argc, char **argv, char *xoptions, char *xtag, char *xdate,
 			   argc, argv, local, which, aflag, CVS_LOCK_READ,
 			   preload_update_dir, 1, repository);
 
-#ifdef SERVER_SUPPORT
-    if (server_active)
-	return err;
-#endif
-
     /* see if we need to sleep before returning to avoid time-stamp races */
-    if (last_register_time)
+    if (!server_active && last_register_time)
     {
 	sleep_past (last_register_time);
     }
@@ -527,16 +542,14 @@ do_update (int argc, char **argv, char *xoptions, char *xtag, char *xdate,
  */
 
 static int
-get_linkinfo_proc( void *callerdat, struct file_info *finfo )
+get_linkinfo_proc (void *callerdat, struct file_info *finfo)
 {
     char *fullpath;
     Node *linkp;
     struct hardlink_info *hlinfo;
 
     /* Get the full pathname of the current file. */
-    fullpath = xmalloc (strlen(working_dir) +
-			strlen(finfo->fullname) + 2);
-    sprintf (fullpath, "%s/%s", working_dir, finfo->fullname);
+    fullpath = Xasprintf ("%s/%s", working_dir, finfo->fullname);
 
     /* To permit recursing into subdirectories, files
        are keyed on the full pathname and not on the basename. */
@@ -549,8 +562,7 @@ get_linkinfo_proc( void *callerdat, struct file_info *finfo )
     }
     
     /* Create a new, empty hardlink_info node. */
-    hlinfo = (struct hardlink_info *)
-	xmalloc (sizeof (struct hardlink_info));
+    hlinfo = xmalloc (sizeof (struct hardlink_info));
 
     hlinfo->status = (Ctype) 0;	/* is this dumb? */
     hlinfo->checked_out = 0;
@@ -672,11 +684,7 @@ update_fileproc (void *callerdat, struct file_info *finfo)
                     bakname = backup_file (finfo->file, vers->vn_user);
                     /* This behavior is sufficiently unexpected to
                        justify overinformativeness, I think. */
-#ifdef SERVER_SUPPORT
-                    if ((! really_quiet) && (! server_active))
-#else /* ! SERVER_SUPPORT */
-                    if (! really_quiet)
-#endif /* SERVER_SUPPORT */
+                    if (!really_quiet && !server_active)
                         (void) printf ("(Locally modified %s moved to %s)\n",
                                        finfo->file, bakname);
                     free (bakname);
@@ -691,8 +699,7 @@ update_fileproc (void *callerdat, struct file_info *finfo)
                 {
                     if (vers->ts_conflict)
                     {
-			if (file_has_conflict (finfo, vers->ts_conflict)
-			    || file_has_markers (finfo))
+			if (file_has_markers (finfo))
                         {
                             write_letter (finfo, 'C');
                             retval = 1;
@@ -703,7 +710,7 @@ update_fileproc (void *callerdat, struct file_info *finfo)
                             Register (finfo->entries, finfo->file, 
                                       vers->vn_rcs, vers->ts_rcs,
                                       vers->options, vers->tag,
-                                      vers->date, (char *)0);
+                                      vers->date, NULL);
                         }
                     }
                     if (!retval)
@@ -729,7 +736,7 @@ update_fileproc (void *callerdat, struct file_info *finfo)
 					     ? SERVER_RCS_DIFF
 					     : SERVER_PATCHED),
 					    file_info.st_mode, checksum,
-					    (struct buffer *) NULL);
+					    NULL);
 			break;
 		    }
 		}
@@ -791,17 +798,11 @@ update_ignproc (const char *file, const char *dir)
     memset (&finfo, 0, sizeof (finfo));
     finfo.file = file;
     finfo.update_dir = dir;
-    if (dir[0] == '\0')
-	tmp = xstrdup (file);
-    else
-    {
-	tmp = xmalloc (strlen (file) + strlen (dir) + 10);
-	strcpy (tmp, dir);
-	strcat (tmp, "/");
-	strcat (tmp, file);
-    }
 
-    finfo.fullname = tmp;
+    finfo.fullname = tmp = Xasprintf ("%s%s%s",
+				      dir[0] == '\0' ? "" : dir,
+				      dir[0] == '\0' ? "" : "/",
+				      file);
     write_letter (&finfo, '?');
     free (tmp);
 }
@@ -836,11 +837,7 @@ update_filesdone_proc (void *callerdat, int err, const char *repository,
 	if (unlink_file_dir (CVSADM) < 0 && !existence_error (errno))
 	    error (0, errno, "cannot remove %s directory", CVSADM);
     }
-#ifdef SERVER_SUPPORT
     else if (!server_active && !pipeout)
-#else
-    else if (!pipeout)
-#endif /* SERVER_SUPPORT */
     {
         /* If there is no CVS/Root file, add one */
         if (!isfile (CVSADM_ROOT))
@@ -889,15 +886,11 @@ update_dirent_proc (void *callerdat, const char *dir, const char *repository,
 	   is when update -d is specified, and the working directory
 	   is gone but the subdirectory is still mentioned in
 	   CVS/Entries).  */
-	if (1
-#ifdef SERVER_SUPPORT
-	    /* In the remote case, the client should refrain from
-	       sending us the directory in the first place.  So we
-	       want to continue to give an error, so clients make
-	       sure to do this.  */
-	    && !server_active
-#endif
-	    && !isdir (repository))
+	/* In the remote case, the client should refrain from
+	   sending us the directory in the first place.  So we
+	   want to continue to give an error, so clients make
+	   sure to do this.  */
+	if (!server_active && !isdir (repository))
 	    return R_SKIP_ALL;
 
 	if (noexec)
@@ -950,10 +943,7 @@ update_dirent_proc (void *callerdat, const char *dir, const char *repository,
 	/* The directory exists.  Check to see if it has a CVS
 	   subdirectory.  */
 
-	cvsadmdir = xmalloc (strlen (dir) + 80);
-	strcpy (cvsadmdir, dir);
-	strcat (cvsadmdir, "/");
-	strcat (cvsadmdir, CVSADM);
+	cvsadmdir = Xasprintf ("%s/%s", dir, CVSADM);
 
 	if (!isdir (cvsadmdir))
 	{
@@ -974,10 +964,8 @@ update_dirent_proc (void *callerdat, const char *dir, const char *repository,
     {
 	if (update_build_dirs)
 	{
-	    char *tmp;
+	    char *tmp = Xasprintf ("%s/%s", dir, CVSADM_ENTSTAT);
 
-	    tmp = xmalloc (strlen (dir) + sizeof (CVSADM_ENTSTAT) + 10);
-	    (void) sprintf (tmp, "%s/%s", dir, CVSADM_ENTSTAT);
 	    if (unlink_file (tmp) < 0 && ! existence_error (errno))
 		error (1, errno, "cannot remove file %s", tmp);
 #ifdef SERVER_SUPPORT
@@ -1177,13 +1165,10 @@ scratch_file (struct file_info *finfo, Vers_TS *vers)
 #endif
     if (unlink_file (finfo->file) < 0 && ! existence_error (errno))
 	error (0, errno, "unable to remove %s", finfo->fullname);
-    else
-#ifdef SERVER_SUPPORT
+    else if (!server_active)
+    {
 	/* skip this step when the server is running since
 	 * server_updated should have handled it */
-	if (!server_active)
-#endif
-    {
 	/* keep the vers structure up to date in case we do a join
 	 * - if there isn't a file, it can't very well have a version number, can it?
 	 */
@@ -1221,17 +1206,9 @@ checkout_file (struct file_info *finfo, Vers_TS *vers_ts, int adding,
 
     /* Don't screw with backup files if we're going to stdout, or if
        we are the server.  */
-    if (!pipeout
-#ifdef SERVER_SUPPORT
-	&& ! server_active
-#endif
-	)
+    if (!pipeout && !server_active)
     {
-	backup = xmalloc (strlen (finfo->file)
-			  + sizeof (CVSADM)
-			  + sizeof (CVSPREFIX)
-			  + 10);
-	(void) sprintf (backup, "%s/%s%s", CVSADM, CVSPREFIX, finfo->file);
+	backup = Xasprintf ("%s/%s%s", CVSADM, CVSPREFIX, finfo->file);
 	if (isfile (finfo->file))
 	    rename_file (finfo->file, backup);
 	else
@@ -1314,7 +1291,7 @@ VERS: ", 0);
 		   That would also fix the kludge with noexec, above, which
 		   is here only because noexec doesn't write srcfile->path
 		   for us to stat.  */
-		if( CVS_STAT( vers_ts->srcfile->path, &sb ) < 0 )
+		if (stat (vers_ts->srcfile->path, &sb) < 0)
 		{
 #if defined (SERVER_SUPPORT) || defined (CLIENT_SUPPORT)
 		    buf_free (revbuf);
@@ -1454,7 +1431,7 @@ VERS: ", 0);
 	if (update_server && server_active)
 	    server_updated (finfo, vers_ts,
 			    merging ? SERVER_MERGED : SERVER_UPDATED,
-			    mode, (unsigned char *) NULL, revbuf);
+			    mode, NULL, revbuf);
 #endif
     }
     else
@@ -1533,7 +1510,8 @@ struct patch_file_data
  * itself.
  */
 static int
-patch_file (struct file_info *finfo, Vers_TS *vers_ts, int *docheckout, struct stat *file_info, unsigned char *checksum)
+patch_file (struct file_info *finfo, Vers_TS *vers_ts, int *docheckout,
+	    struct stat *file_info, unsigned char *checksum)
 {
     char *backup;
     char *file1;
@@ -1589,11 +1567,7 @@ patch_file (struct file_info *finfo, Vers_TS *vers_ts, int *docheckout, struct s
 	return 0;
     }
 
-    backup = xmalloc (strlen (finfo->file)
-		      + sizeof (CVSADM)
-		      + sizeof (CVSPREFIX)
-		      + 10);
-    (void) sprintf (backup, "%s/%s%s", CVSADM, CVSPREFIX, finfo->file);
+    backup = Xasprintf ("%s/%s%s", CVSADM, CVSPREFIX, finfo->file);
     if (isfile (finfo->file))
         rename_file (finfo->file, backup);
     else
@@ -1603,16 +1577,8 @@ patch_file (struct file_info *finfo, Vers_TS *vers_ts, int *docheckout, struct s
 	    error (0, errno, "cannot remove %s", backup);
     }
 
-    file1 = xmalloc (strlen (finfo->file)
-		     + sizeof (CVSADM)
-		     + sizeof (CVSPREFIX)
-		     + 10);
-    (void) sprintf (file1, "%s/%s%s-1", CVSADM, CVSPREFIX, finfo->file);
-    file2 = xmalloc (strlen (finfo->file)
-		     + sizeof (CVSADM)
-		     + sizeof (CVSPREFIX)
-		     + 10);
-    (void) sprintf (file2, "%s/%s%s-2", CVSADM, CVSPREFIX, finfo->file);
+    file1 = Xasprintf ("%s/%s%s-1", CVSADM, CVSPREFIX, finfo->file);
+    file2 = Xasprintf ("%s/%s%s-2", CVSADM, CVSPREFIX, finfo->file);
 
     fail = 0;
 
@@ -1641,7 +1607,7 @@ patch_file (struct file_info *finfo, Vers_TS *vers_ts, int *docheckout, struct s
      * At least, if we are keeping track of the tag vn_user came from,
      * I don't know where yet. -DRP
      */
-    retcode = RCS_checkout (vers_ts->srcfile, (char *) NULL,
+    retcode = RCS_checkout (vers_ts->srcfile, NULL,
 			    vers_ts->vn_user, vers_ts->tag,
 			    vers_ts->options, RUN_TTY,
 			    patch_file_write, (void *) &data);
@@ -1664,7 +1630,7 @@ patch_file (struct file_info *finfo, Vers_TS *vers_ts, int *docheckout, struct s
 	data.compute_checksum = 1;
 	md5_init_ctx (&data.context);
 
-	retcode = RCS_checkout (vers_ts->srcfile, (char *) NULL,
+	retcode = RCS_checkout (vers_ts->srcfile, NULL,
 				vers_ts->vn_rcs, vers_ts->tag,
 				vers_ts->options, RUN_TTY,
 				patch_file_write, (void *) &data);
@@ -1681,7 +1647,9 @@ patch_file (struct file_info *finfo, Vers_TS *vers_ts, int *docheckout, struct s
     retcode = 0;
     if (! fail)
     {
-	char *diff_options;
+	int dargc = 0;
+	size_t darg_allocated = 0;
+	char **dargv = NULL;
 
 	/* If the client does not support the Rcs-diff command, we
            send a context diff, and the client must invoke patch.
@@ -1689,16 +1657,13 @@ patch_file (struct file_info *finfo, Vers_TS *vers_ts, int *docheckout, struct s
            new approach only requires running diff in the server; the
            client can handle everything without invoking an external
            program.  */
-	if (! rcs_diff_patches)
-	{
+	if (!rcs_diff_patches)
 	    /* We use -c, not -u, because that is what CVS has
 	       traditionally used.  Kind of a moot point, now that
 	       Rcs-diff is preferred, so there is no point in making
 	       the compatibility issues worse.  */
-	    diff_options = "-c";
-	}
+	    run_add_arg_p (&dargc, &darg_allocated, &dargv, "-c");
 	else
-	{
 	    /* Now that diff is librarified, we could be passing -a if
 	       we wanted to.  However, it is unclear to me whether we
 	       would want to.  Does diff -a, in any significant
@@ -1708,27 +1673,25 @@ patch_file (struct file_info *finfo, Vers_TS *vers_ts, int *docheckout, struct s
 	       'binary'.  Conversely, do they tend to be much larger
 	       in the bad cases?  This needs some more
 	       thought/investigation, I suspect.  */
-
-	    diff_options = "-n";
-	}
-	retcode = diff_exec (file1, file2, NULL, NULL, diff_options, finfo->file);
+	    run_add_arg_p (&dargc, &darg_allocated, &dargv, "-n");
+	retcode = diff_exec (file1, file2, NULL, NULL, dargc, dargv,
+			     finfo->file);
+	run_arg_free_p (dargc, dargv);
+	free (dargv);
 
 	/* A retcode of 0 means no differences.  1 means some differences.  */
-	if (retcode != 0
-	    && retcode != 1)
-	{
+	if (retcode != 0 && retcode != 1)
 	    fail = 1;
-	}
     }
 
-    if (! fail)
+    if (!fail)
     {
 	struct stat file2_info;
 
 	/* Check to make sure the patch is really shorter */
-	if (CVS_STAT (file2, &file2_info) < 0)
+	if (stat (file2, &file2_info) < 0)
 	    error (1, errno, "could not stat %s", file2);
-	if (CVS_STAT (finfo->file, file_info) < 0)
+	if (stat (finfo->file, file_info) < 0)
 	    error (1, errno, "could not stat %s", finfo->file);
 	if (file2_info.st_size <= file_info->st_size)
 	    fail = 1;
@@ -1763,7 +1726,7 @@ patch_file (struct file_info *finfo, Vers_TS *vers_ts, int *docheckout, struct s
 	/* Stat the original RCS file, and then adjust it the way
 	   that RCS_checkout would.  FIXME: This is an abstraction
 	   violation.  */
-	if (CVS_STAT (vers_ts->srcfile->path, file_info) < 0)
+	if (stat (vers_ts->srcfile->path, file_info) < 0)
 	    error (1, errno, "could not stat %s", vers_ts->srcfile->path);
 	if (chmod (finfo->file,
 		   file_info->st_mode & ~(S_IWRITE | S_IWGRP | S_IWOTH))
@@ -1784,7 +1747,7 @@ patch_file (struct file_info *finfo, Vers_TS *vers_ts, int *docheckout, struct s
 		  xvers_ts->ts_user, xvers_ts->options,
 		  xvers_ts->tag, xvers_ts->date, NULL);
 
-	if (CVS_STAT (finfo->file, file_info) < 0)
+	if (stat (finfo->file, file_info) < 0)
 	    error (1, errno, "could not stat %s", finfo->file);
 
 	/* If this is really Update and not Checkout, record history.  */
@@ -1897,6 +1860,41 @@ write_letter (struct file_info *finfo, int letter)
 
 
 
+/* Reregister a file after a merge.  */
+static void
+RegisterMerge (struct file_info *finfo, Vers_TS *vers,
+	       const char *backup, int has_conflicts)
+{
+    /* This file is the result of a merge, which means that it has
+       been modified.  We use a special timestamp string which will
+       not compare equal to any actual timestamp.  */
+    char *cp = NULL;
+
+    if (has_conflicts)
+    {
+	time (&last_register_time);
+	cp = time_stamp (finfo->file);
+    }
+    Register (finfo->entries, finfo->file, vers->vn_rcs ? vers->vn_rcs : "0",
+	      "Result of merge", vers->options, vers->tag, vers->date, cp);
+    if (cp)
+	free (cp);
+
+#ifdef SERVER_SUPPORT
+    /* Send the new contents of the file before the message.  If we
+       wanted to be totally correct, we would have the client write
+       the message only after the file has safely been written.  */
+    if (server_active)
+    {
+        server_copy_file (finfo->file, finfo->update_dir, finfo->repository,
+			  backup);
+	server_updated (finfo, vers, SERVER_MERGED, (mode_t) -1, NULL, NULL);
+    }
+#endif
+}
+
+
+
 /*
  * Do all the magic associated with a file which needs to be merged
  */
@@ -1907,6 +1905,8 @@ merge_file (struct file_info *finfo, Vers_TS *vers)
     int status;
     int retval;
 
+    assert (vers->vn_user);
+
     /*
      * The users currently modified file is moved to a backup file name
      * ".#filename.version", so that it will stay around for a few days
@@ -1914,11 +1914,7 @@ merge_file (struct file_info *finfo, Vers_TS *vers)
      * is the version of the file that the user was most up-to-date with
      * before the merge.
      */
-    backup = xmalloc (strlen (finfo->file)
-		      + strlen (vers->vn_user)
-		      + sizeof (BAKPREFIX)
-		      + 10);
-    (void) sprintf (backup, "%s%s.%s", BAKPREFIX, finfo->file, vers->vn_user);
+    backup = Xasprintf ("%s%s.%s", BAKPREFIX, finfo->file, vers->vn_user);
 
     if (unlink_file (backup) && !existence_error (errno))
 	error (0, errno, "unable to remove %s", backup);
@@ -1944,13 +1940,21 @@ merge_file (struct file_info *finfo, Vers_TS *vers)
 	   thought needs to go into this, and in the meantime it is safe
 	   to treat any such mismatch as an automatic conflict. -twp */
 
-#ifdef SERVER_SUPPORT
-	if (server_active)
-	    server_copy_file (finfo->file, finfo->update_dir,
-			      finfo->repository, backup);
-#endif
+	status = RCS_checkout (finfo->rcs, finfo->file, vers->vn_rcs,
+			       vers->tag, vers->options, NULL, NULL, NULL);
+	if (status)
+	{
+	    error (0, 0, "failed to check out `%s' file", finfo->fullname);
+	    error (0, 0, "restoring `%s' from backup file `%s'",
+		   finfo->fullname, backup);
+	    rename_file (backup, finfo->file);
+	    retval = 1;
+	    goto out;
+	}
 
-	status = checkout_file (finfo, vers, 0, 1, 1);
+	xchmod (finfo->file, 1);
+
+	RegisterMerge (finfo, vers, backup, 1);
 
 	/* Is there a better term than "nonmergeable file"?  What we
 	   really mean is, not something that CVS cannot or does not
@@ -1984,24 +1988,6 @@ merge_file (struct file_info *finfo, Vers_TS *vers)
     if (strcmp (vers->options, "-V4") == 0)
 	vers->options[0] = '\0';
 
-    /* This file is the result of a merge, which means that it has
-       been modified.  We use a special timestamp string which will
-       not compare equal to any actual timestamp.  */
-    {
-	char *cp = 0;
-
-	if (status)
-	{
-	    (void) time (&last_register_time);
-	    cp = time_stamp (finfo->file);
-	}
-	Register (finfo->entries, finfo->file, vers->vn_rcs,
-		  "Result of merge", vers->options, vers->tag,
-		  vers->date, cp);
-	if (cp)
-	    free (cp);
-    }
-
     /* fix up the vers structure, in case it is used by join */
     if (join_rev1)
     {
@@ -2012,19 +1998,7 @@ merge_file (struct file_info *finfo, Vers_TS *vers)
 	vers->vn_user = xstrdup (vers->vn_rcs);
     }
 
-#ifdef SERVER_SUPPORT
-    /* Send the new contents of the file before the message.  If we
-       wanted to be totally correct, we would have the client write
-       the message only after the file has safely been written.  */
-    if (server_active)
-    {
-        server_copy_file (finfo->file, finfo->update_dir, finfo->repository,
-			  backup);
-	server_updated (finfo, vers, SERVER_MERGED,
-			(mode_t) -1, (unsigned char *) NULL,
-			(struct buffer *) NULL);
-    }
-#endif
+    RegisterMerge (finfo, vers, backup, status);
 
     if (status == 1)
     {
@@ -2096,19 +2070,19 @@ join_file (struct file_info *finfo, Vers_TS *vers)
     char *jdate1;
     char *jdate2;
 
-    TRACE ( 1, "join_file(%s, %s%s%s%s, %s, %s)",
-	    finfo->file,
-	    vers->tag ? vers->tag : "",
-	    vers->tag ? " (" : "",
-	    vers->vn_rcs ? vers->vn_rcs : "",
-	    vers->tag ? ")" : "",
-	    join_rev1 ? join_rev1 : "",
-	    join_rev2 ? join_rev2 : "" );
+    TRACE (TRACE_FUNCTION, "join_file(%s, %s%s%s%s, %s, %s)",
+	   finfo->file,
+	   vers->tag ? vers->tag : "",
+	   vers->tag ? " (" : "",
+	   vers->vn_rcs ? vers->vn_rcs : "",
+	   vers->tag ? ")" : "",
+	   join_rev1 ? join_rev1 : "",
+	   join_rev2 ? join_rev2 : "");
 
     jrev1 = join_rev1;
     jrev2 = join_rev2;
-    jdate1 = date_rev1;
-    jdate2 = date_rev2;
+    jdate1 = join_date1;
+    jdate2 = join_date2;
 
     /* Determine if we need to do anything at all.  */
     if (vers->srcfile == NULL ||
@@ -2131,14 +2105,14 @@ join_file (struct file_info *finfo, Vers_TS *vers)
        below about vn_user.  */
 
     /* Convert the second revision, walking branches and dates.  */
-    rev2 = RCS_getversion (vers->srcfile, jrev2, jdate2, 1, (int *) NULL);
+    rev2 = RCS_getversion (vers->srcfile, jrev2, jdate2, 1, NULL);
 
     /* If this is a merge of two revisions, get the first revision.
        If only one join tag was specified, then the first revision is
        the greatest common ancestor of the second revision and the
        working file.  */
     if (jrev1 != NULL)
-	rev1 = RCS_getversion (vers->srcfile, jrev1, jdate1, 1, (int *) NULL);
+	rev1 = RCS_getversion (vers->srcfile, jrev1, jdate1, 1, NULL);
     else
     {
 	/* Note that we use vn_rcs here, since vn_user may contain a
@@ -2282,14 +2256,13 @@ join_file (struct file_info *finfo, Vers_TS *vers)
 	   cvs up -rbr -jbr2 could remove and readd the same file
 	 */
 	/* save the rev since server_updated might invalidate it */
-	mrev = xmalloc (strlen (vers->vn_user) + 2);
-	sprintf (mrev, "-%s", vers->vn_user);
+	mrev = Xasprintf ("-%s", vers->vn_user);
 #ifdef SERVER_SUPPORT
 	if (server_active)
 	{
 	    server_scratch (finfo->file);
 	    server_updated (finfo, vers, SERVER_UPDATED, (mode_t) -1,
-			    (unsigned char *) NULL, (struct buffer *) NULL);
+			    NULL, NULL);
 	}
 #endif
 	Register (finfo->entries, finfo->file, mrev, vers->ts_rcs,
@@ -2430,8 +2403,7 @@ join_file (struct file_info *finfo, Vers_TS *vers)
 	 */
 	retcode = RCS_checkout (vers->srcfile, finfo->file,
 				vers->vn_user, vers->tag,
-				(char *) NULL, RUN_TTY,
-				(RCSCHECKOUTPROC) NULL, (void *) NULL);
+				NULL, RUN_TTY, NULL, NULL);
 	if (retcode != 0)
 	    error (1, 0,
 		   "failed to check out %s file", finfo->fullname);
@@ -2445,11 +2417,7 @@ join_file (struct file_info *finfo, Vers_TS *vers)
      * is the version of the file that the user was most up-to-date with
      * before the merge.
      */
-    backup = xmalloc (strlen (finfo->file)
-		      + strlen (vers->vn_user)
-		      + sizeof (BAKPREFIX)
-		      + 10);
-    (void) sprintf (backup, "%s%s.%s", BAKPREFIX, finfo->file, vers->vn_user);
+    backup = Xasprintf ("%s%s.%s", BAKPREFIX, finfo->file, vers->vn_user);
 
     if (unlink_file (backup) < 0
 	&& !existence_error (errno))
@@ -2496,8 +2464,8 @@ join_file (struct file_info *finfo, Vers_TS *vers)
 	 * Also, it is safe to pass in NULL for nametag since we know no
 	 * substitution is happening during the binary mode checkout.
 	 */
-	if (RCS_checkout ( finfo->rcs, finfo->file, rev2, (char *)NULL, t_options,
-			   RUN_TTY, (RCSCHECKOUTPROC)0, NULL) != 0 )
+	if (RCS_checkout (finfo->rcs, finfo->file, rev2, NULL, t_options,
+			  RUN_TTY, NULL, NULL) != 0)
 	    status = 2;
 	else
 	    status = 0;
@@ -2531,8 +2499,8 @@ join_file (struct file_info *finfo, Vers_TS *vers)
 	   the two files, and let them resolve it.  It is possible
 	   that we should require a "touch foo" or similar step before
 	   we allow a checkin.  */
-	if (RCS_checkout ( finfo->rcs, finfo->file, rev2, (char *)NULL,
-			   t_options, RUN_TTY, (RCSCHECKOUTPROC)0, NULL) != 0)
+	if (RCS_checkout (finfo->rcs, finfo->file, rev2, NULL,
+			  t_options, RUN_TTY, NULL, NULL) != 0)
 	    status = 2;
 	else
 	    status = 0;
@@ -2608,31 +2576,7 @@ join_file (struct file_info *finfo, Vers_TS *vers)
        RCS_checkout above, and we aren't running as the server.
        However, that is not the normal case, and calling Register
        again won't cost much in that case.  */
-    {
-	char *cp = 0;
-
-	if (status)
-	{
-	    (void) time (&last_register_time);
-	    cp = time_stamp (finfo->file);
-	}
-	Register (finfo->entries, finfo->file,
-		  vers->vn_rcs ? vers->vn_rcs : "0", "Result of merge",
-		  vers->options, vers->tag, vers->date, cp);
-	if (cp)
-	    free(cp);
-    }
-
-#ifdef SERVER_SUPPORT
-    if (server_active)
-    {
-	server_copy_file (finfo->file, finfo->update_dir, finfo->repository,
-			  backup);
-	server_updated (finfo, vers, SERVER_MERGED,
-			(mode_t) -1, (unsigned char *) NULL,
-			(struct buffer *) NULL);
-    }
-#endif
+    RegisterMerge (finfo, vers, backup, status);
 
 out:
     free (rev1);
@@ -2702,7 +2646,7 @@ special_file_mismatch (struct file_info *finfo, char *rev1, char *rev2)
 	else
 	{
 # ifdef HAVE_STRUCT_STAT_ST_RDEV
-	    if (CVS_LSTAT (finfo->file, &sb) < 0)
+	    if (lstat (finfo->file, &sb) < 0)
 		error (1, errno, "could not get file information for %s",
 		       finfo->file);
 	    rev1_uid = sb.st_uid;
@@ -2782,7 +2726,7 @@ special_file_mismatch (struct file_info *finfo, char *rev1, char *rev2)
 	else
 	{
 # ifdef HAVE_STRUCT_STAT_ST_RDEV
-	    if (CVS_LSTAT (finfo->file, &sb) < 0)
+	    if (lstat (finfo->file, &sb) < 0)
 		error (1, errno, "could not get file information for %s",
 		       finfo->file);
 	    rev2_uid = sb.st_uid;
@@ -2959,5 +2903,5 @@ special_file_mismatch (struct file_info *finfo, char *rev1, char *rev2)
 int
 joining (void)
 {
-    return join_rev1 != NULL;
+    return join_rev1 || join_date1;
 }
